from typing import Any, Dict, List, Optional, Type, Union, Iterator
import os
import platform
import time
import logging
import socket

import docker
from pydantic import BaseModel
import requests
from docker.api.client import APIClient
from docker.errors import NotFound, APIError
from tqdm import tqdm
from namesgenerator import get_random_name

from .base import DesktopInstance, V1ProviderData, DesktopProvider
from agentdesk.util import find_open_port


logger = logging.getLogger(__name__)


class DockerConnectConfig(BaseModel):
    timeout: Optional[int] = None


class DockerProvider(DesktopProvider):
    """A docker desktop provider"""

    def __init__(self, cfg: Optional[DockerConnectConfig] = None) -> None:
        self._configure_docker_socket()
        if not cfg:
            cfg = DockerConnectConfig()

        self._cfg = cfg
        if cfg.timeout:
            self.client = docker.from_env(timeout=cfg.timeout)
        else:
            self.client = docker.from_env()

    def _configure_docker_socket(self):
        if os.path.exists("/var/run/docker.sock"):
            docker_socket = "unix:///var/run/docker.sock"
        else:
            user = os.environ.get("USER")
            if os.path.exists(f"/Users/{user}/.docker/run/docker.sock"):
                docker_socket = f"unix:///Users/{user}/.docker/run/docker.sock"
            else:
                raise FileNotFoundError(
                    (
                        "Neither '/var/run/docker.sock' nor '/Users/<USER>/.docker/run/docker.sock' are available."
                        "Please make sure you have Docker installed and running."
                    )
                )
        os.environ["DOCKER_HOST"] = docker_socket

    def create(
        self,
        name: Optional[str] = None,
        image: Optional[str] = None,
        memory: int = 4,
        cpu: int = 2,
        disk: str = "30gb",
        tags: Optional[Dict[str, str]] = None,
        reserve_ip: bool = False,
        ssh_key_pair: Optional[str] = None,
        owner_id: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        enable_basic_auth: bool = False,
        password: Optional[str] = None,
        sub_folder: Optional[str] = None,
        id: Optional[str] = None,
        ttl: Optional[int] = None,
    ) -> DesktopInstance:
        """Create a Desktop

        Args:
            name (str, optional): Name of the instance. Defaults to random generation.
            image (str, optional): Image of the instance. Defaults to Ubuntu Jammy.
            memory (int): Memory allotment. Defaults to 4gb.
            cpu (int): CPU allotment. Defaults to 2.
            disk (str): Disk allotment. Defaults to 30gb.
            tags (List[str], optional): Tags to apply to the instance. Defaults to None.
            reserve_ip (bool, optional): Reserve an IP address. Defaults to False.
            ssh_key_pair (str, optional): SSH key pair name to use. Defaults to None.
            owner_id (str, optional): Owner of the instance. Defaults to None.
            metadata (Dict[str, Any], optional): Metadata to apply to the intance. Defaults to None.
            enable_basic_auth (bool, optional): Enable basic auth for the instance. Defaults to False.
            password (str, optional): Password to use for basic auth. Defaults to None.
            sub_folder (str, optional): Subfolder to use. Defaults to None.
            id (str, optional): ID of the instance. Defaults to None.
            ttl (int, optional): Time to live for the instance. Defaults to None.

        Returns:
            DesktopInstance: A desktop instance
        """

        if reserve_ip:
            raise NotImplementedError(
                "Reserving IP addresses is not supported for Docker"
            )
        if ssh_key_pair:
            raise NotImplementedError("SSH key pairs are not supported for Docker")
        if enable_basic_auth:
            raise NotImplementedError("Basic auth is not supported for Docker")
        if password:
            raise ValueError("Password is not supported for Docker")

        if not name:
            name = get_random_name("-")
            if not name:
                raise ValueError("Could not generate a random name")

        labels = {
            "provisioner": "agentdesk",
            "desktop_name": name,
        }

        if not image:
            image = "us-docker.pkg.dev/agentsea-dev/agentd/desktop-webtop:latest"

        agentd_port = find_open_port(8000, 9000)
        if not agentd_port:
            raise ValueError("Could not find open port")

        vnc_port = find_open_port(3000, 4000)
        if not vnc_port:
            raise ValueError("Could not find open port")

        vnc_port_https = find_open_port(3100, 4100)
        if not vnc_port_https:
            raise ValueError("Could not find open port")

        env_vars = {}
        if sub_folder:
            env_vars["SUBFOLDER"] = sub_folder

        # Initialize tqdm progress bar
        api_client = APIClient()

        # Pull the image with progress tracking
        pull_image(image, api_client)

        print(f"running image {image}")
        self.ensure_network("agentsea")

        container_params = {
            "image": image,
            "network": "agentsea",
            "ports": {
                8000: agentd_port,
                3000: vnc_port,
                3001: vnc_port_https,
            },
            "environment": env_vars,
            "detach": True,
            "labels": labels,
            "name": name,
        }

        # Add extra_hosts only for Linux
        if platform.system() == "Linux":
            container_params["extra_hosts"] = {"host.docker.internal": "host-gateway"}

        try:
            container = self.client.containers.run(**container_params)
        except Exception as e:
            raise RuntimeError(f"Could not run docker desktop container '{name}': {e}")
        if container and type(container) != bytes:
            print(f"container id '{container.id}'")  # type: ignore

        # Wait for the container to be in the "running" state
        for _ in range(10):
            container.reload()  # type: ignore
            if container.status == "running":  # type: ignore
                break
            time.sleep(1)
        else:
            raise RuntimeError(f"Container '{name}' did not start in time")

        # Check /health endpoint
        health_url = f"http://localhost:{agentd_port}/health"
        for _ in range(60):
            try:
                print("waiting for agent to be ready...")
                print(f"checking health at {health_url}")
                response = requests.get(health_url)
                print(f"response: {response}")
                if response.status_code == 200:
                    print(f"Health check passed for '{name}'")
                    break
            except requests.RequestException as e:
                logger.debug(f"Health check failed: {e}")
            time.sleep(1)
        else:
            container.remove(force=True)  # type: ignore
            raise RuntimeError(f"Container '{name}' did not pass health check")

        return DesktopInstance(
            id=id,
            name=name,
            addr=f"http://localhost:{agentd_port}",
            cpu=cpu,
            memory=memory,
            disk=disk,
            owner_id=owner_id,
            metadata=metadata,
            agentd_port=agentd_port,
            vnc_port=vnc_port,
            vnc_port_https=vnc_port_https,
            requires_proxy=False,
            provider=self.to_data(),
            ttl=ttl,
        )

    def delete(self, name: str, owner_id: Optional[str] = None) -> None:
        """Delete a VM

        Args:
            name (str): Name of the VM
            owner_id (str, optional): Owner of the VM. Defaults to None
        """
        try:
            # Attempt to get the container by name
            container = self.client.containers.get(name)

            # If found, remove the container
            container.remove(force=True)  # type: ignore
            print(f"Successfully deleted container: {name}")
        except NotFound:
            # Handle the case where the container does not exist
            print(f"Container '{name}' does not exist.")
            raise
        except Exception as e:
            # Handle other potential errors
            print(f"Failed to delete container '{name}': {e}")
            raise

    def start(
        self,
        name: str,
        private_ssh_key: Optional[str] = None,
        owner_id: Optional[str] = None,
    ) -> None:
        """Start a VM

        Args:
            name (str): Name of the VM
            private_ssh_key (str, optional): SSH key to use. Defaults to use ~/.ssh/id_rsa.
        """
        raise NotImplementedError("Not implemented for container runtimes")

    def stop(self, name: str, owner_id: Optional[str] = None) -> None:
        """Stop a VM

        Args:
            name (str): Name of the VM
            owner_id (str, optional): Owner of the VM. Defaults to None
        """
        raise NotImplementedError("Not implemented for container runtimes")

    def list(self) -> List[DesktopInstance]:
        """List VMs

        Returns:
            List[VM]: A list of VMs
        """
        instances = DesktopInstance.find()

        out = []

        for instance in instances:
            if not instance.provider:
                continue

            if instance.provider.type == "docker":
                out.append(instance)

        return out

    def get(
        self, name: str, owner_id: Optional[str] = None
    ) -> Optional[DesktopInstance]:
        """Get a VM

        Args:
            name (str): Name of the VM
            owner_id (str, optional): Owner of the VM. Defaults to None
        """
        instances = DesktopInstance.find(name=name, owner_id=owner_id)
        if len(instances) == 0:
            return None
        return instances[0]

    def to_data(self) -> V1ProviderData:
        """Convert to a ProviderData object

        Returns:
            ProviderData: ProviderData object
        """
        args = {}
        if self._cfg:
            args["cfg"] = self._cfg.model_dump_json()

        return V1ProviderData(
            type="docker",
            args=args,
        )

    @classmethod
    def from_data(cls, data: V1ProviderData) -> "DockerProvider":
        """From provider data

        Args:
            data (ProviderData): Provider data
        """
        if data.args:
            cfg = DockerConnectConfig.model_validate_json(data.args["cfg"])
        else:
            cfg = DockerConnectConfig()
        return cls(cfg=cfg)

    def refresh(self, log: bool = True) -> None:
        """Refresh state"""
        label_filter = {"label": "provisioner=agentdesk"}
        running_containers = self.client.containers.list(filters=label_filter)

        # Fetch the agent instances from the database
        db_instances = self.list()

        # Create a mapping of container names to containers
        running_containers_map = {
            container.name: container for container in running_containers
        }  # type: ignore

        # Create a mapping of instance names to instances
        db_instances_map = {instance.name: instance for instance in db_instances}

        # Check for instances in the database that are not running as containers
        for instance_name, instance in db_instances_map.items():
            if instance_name not in running_containers_map:
                print(
                    f"Instance '{instance_name}' is in the database but not running. Removing from database."
                )
                instance.delete(force=True)

        logger.debug(
            "Refresh complete. State synchronized between Docker and the database."
        )

    @classmethod
    def name(cls) -> str:
        return "docker"

    @classmethod
    def connect_config_type(cls) -> Type[DockerConnectConfig]:
        return DockerConnectConfig

    def connect_config(self) -> DockerConnectConfig:
        return self._cfg

    @classmethod
    def connect(cls, cfg: DockerConnectConfig) -> "DockerProvider":
        return cls(cfg)

    def ensure_network(self, network_name: str) -> None:
        """Ensure that the specified Docker network exists, creating it if necessary."""
        try:
            self.client.networks.get(network_name)
            print(f"Network '{network_name}' already exists.")
        except NotFound:
            self.client.networks.create(network_name)
            print(f"Network '{network_name}' created.")

    def _get_host_ip(self) -> str:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # Doesn't even have to be reachable
            s.connect(("10.254.254.254", 1))
            ip = s.getsockname()[0]
        except Exception:
            ip = "127.0.0.1"
        finally:
            s.close()
        return ip

    def requires_proxy(self) -> bool:
        """Whether this runtime requires a proxy to be used"""
        return False

    def clean(self, owner_id: Optional[str] = None) -> None:
        # Define the filter for containers with the specific label
        label_filter = {"label": ["provisioner=agentdesk"]}

        # Use the filter to list containers
        containers = self.client.containers.list(filters=label_filter, all=True)

        # Initialize a list to keep track of deleted container names or IDs
        deleted_containers = []

        for container in containers:
            try:
                container_name_or_id = (
                    container.name  # type: ignore
                )  # or container.id for container ID
                container.remove(force=True)  # type: ignore
                print(f"Deleted container: {container_name_or_id}")
                deleted_containers.append(container_name_or_id)
            except Exception as e:
                print(f"Failed to delete container: {e}")

        return None

    def logs(
        self, name: str, follow: bool = False, owner_id: Optional[str] = None
    ) -> Union[str, Iterator[str]]:
        """
        Fetches the logs from the specified container. Can return all logs as a single string,
        or stream the logs as a generator of strings.

        Parameters:
            name (str): The name of the container.
            follow (bool): Whether to continuously follow the logs.

        Returns:
            Union[str, Iterator[str]]: All logs as a single string, or a generator that yields log lines.
        """
        try:
            container = self.client.containers.get(name)
            if follow:
                log_stream = container.logs(stream=True, follow=True)  # type: ignore
                return (line.decode("utf-8").strip() for line in log_stream)  # type: ignore
            else:
                return container.logs().decode("utf-8")  # type: ignore
        except NotFound:
            print(f"Container '{name}' does not exist.")
            raise
        except Exception as e:
            print(f"Failed to fetch logs for container '{name}': {e}")
            raise


def pull_image(img: str, api_client: APIClient):
    """
    Pulls a Docker image with progress bars for each layer.

    Args:
        img (str): The Docker image to pull.
        api_client (APIClient): The Docker API client.
    """

    print(f"Pulling Docker image '{img}'...")

    progress_bars = {}
    layers = {}

    try:
        for line in api_client.pull(img, stream=True, decode=True):
            if "id" in line and "progressDetail" in line:
                layer_id = line["id"]
                progress_detail = line["progressDetail"]
                current = progress_detail.get("current", 0)
                total = progress_detail.get("total", 0)

                if total:
                    if layer_id not in layers:
                        progress_bars[layer_id] = tqdm(
                            total=total,
                            desc=f"Layer {layer_id}",
                            leave=False,
                            ncols=100,
                        )
                        layers[layer_id] = 0

                    layers[layer_id] = current
                    progress_bars[layer_id].n = current
                    progress_bars[layer_id].refresh()
            elif "status" in line and "id" in line:
                print(f"Status update for {line['id']}: {line['status']}")
            elif "error" in line:
                raise APIError(line["error"])

    except APIError as e:
        print(f"Error pulling Docker image: {e.explanation}")
    except Exception as e:
        print(f"An unexpected error occurred: {str(e)}")
    finally:
        # Close all progress bars
        for bar in progress_bars.values():
            bar.n = bar.total  # Ensure the progress bar is full before closing
            bar.refresh()
            bar.close()

        print("")



from typing import List, Optional

from pydantic import BaseModel


class V1Health(BaseModel):
    status: Optional[str] = None


class V1Info(BaseModel):
    version: Optional[str] = None


class V1DesktopReqeust(BaseModel):
    name: Optional[str] = None
    memory: Optional[str] = None
    cpu: Optional[int] = None
    disk: Optional[str] = None


class V1DesktopRegistration(BaseModel):
    name: Optional[str] = None
    addr: Optional[str] = None


class V1ProviderData(BaseModel):
    type: Optional[str] = None
    args: Optional[dict] = None


class V1DesktopInstance(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    addr: Optional[str] = None
    status: Optional[str] = None
    created: Optional[float] = None
    assigned: Optional[float] = None
    image: Optional[str] = None
    memory: Optional[int] = None
    cpu: Optional[int] = None
    disk: Optional[str] = None
    memory_usage: Optional[float] = None
    cpu_usage: Optional[float] = None
    disk_usage: Optional[float] = None
    reserved_ip: Optional[bool] = None
    provider: Optional[V1ProviderData] = None
    meta: Optional[dict] = None
    owner_id: Optional[str] = None
    key_pair_name: Optional[str] = None
    agentd_port: Optional[int] = None
    vnc_port: Optional[int] = None
    vnc_port_https: Optional[int] = None
    basic_auth_user: Optional[str] = None
    basic_auth_password: Optional[str] = None
    resource_name: Optional[str] = None
    namespace: Optional[str] = None
    ttl: Optional[int] = None
    requires_proxy: Optional[bool] = None


class V1Desktops(BaseModel):
    desktops: List[V1DesktopInstance]


class V1SSHKey(BaseModel):
    name: str
    public_key: str
    created: float
    id: str
    private_key: Optional[str] = None



from __future__ import annotations
from typing import List, Optional, Dict, Any
import atexit
import time
import logging

import boto3
from mypy_boto3_ec2.service_resource import Instance as EC2Instance
from mypy_boto3_ec2 import EC2Client, EC2ServiceResource
from namesgenerator import get_random_name
from botocore.exceptions import ClientError
import requests

from .base import DesktopInstance, DesktopProvider
from .img import JAMMY
from agentdesk.server.models import V1ProviderData
from agentdesk.util import find_open_port, generate_short_hash, generate_random_string
from agentdesk.proxy import ensure_ssh_proxy, cleanup_proxy
from agentdesk.key import SSHKeyPair


logger = logging.getLogger(__name__)


class EC2Provider(DesktopProvider):
    """VM provider using AWS EC2"""

    AVAILABLE_REGIONS = {
        "us-east-1",
        "us-west-1",
        "us-west-2",
        "eu-west-1",
        "eu-central-1",
        "ap-southeast-1",
        "ap-northeast-1",
    }

    def __init__(
        self,
        region: str,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
    ):
        """Initialize the AWS EC2 VM Provider with region and optional credentials."""
        self.region = region
        if aws_access_key_id and aws_secret_access_key:
            self.session = boto3.Session(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                region_name=self.region,
            )
        else:
            self.session = boto3.Session(region_name=self.region)

        self.ec2: EC2ServiceResource = self.session.resource(
            "ec2", region_name=self.region
        )  # type: ignore
        self.ec2_client: EC2Client = self.session.client("ec2", region_name=self.region)  # type: ignore

    def create(
        self,
        name: Optional[str] = None,
        image: Optional[str] = None,
        memory: int = 4,
        cpu: int = 2,
        disk: str = "30gb",
        tags: Optional[Dict[str, str]] = None,
        reserve_ip: bool = False,
        ssh_key_pair: Optional[str] = None,
        owner_id: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        enable_basic_auth: bool = False,
        password: Optional[str] = None,
        sub_folder: Optional[str] = None,
        id: Optional[str] = None,
        ttl: Optional[int] = None,
    ) -> DesktopInstance:
        if id:
            raise ValueError("cannot set id for ec2 provider")

        if enable_basic_auth:
            raise NotImplementedError(
                "enable_basic_auth not implemented for ec2 provider"
            )
        if password:
            raise NotImplementedError("password not implemented for ec2 provider")
        if sub_folder:
            raise NotImplementedError("sub_folder not implemented for ec2 provider")
        if not name:
            name = get_random_name(sep="-")
            if not name:
                raise ValueError("could not generate name")

        if DesktopInstance.name_exists(name):
            raise ValueError(f"VM name '{name}' already exists")

        if not image:
            image = self._get_ami_id_by_name(JAMMY.ec2)  # type: ignore

        if not ssh_key_pair:
            key_pair = SSHKeyPair.generate_key(
                f"{name}-{generate_short_hash(generate_random_string())}",
                owner_id or "local",
                metadata={"generated_for": name},
            )
            public_ssh_key = key_pair.public_key
            private_ssh_key = key_pair.decrypt_private_key(key_pair.private_key)
        else:
            key_pairs = SSHKeyPair.find(name=ssh_key_pair, owner_id=owner_id or "local")
            if not key_pairs:
                raise ValueError(f"SSH key pair '{ssh_key_pair}' not found")
            key_pair = key_pairs[0]

        public_ssh_key = key_pair.public_key
        private_ssh_key = key_pair.decrypt_private_key(key_pair.private_key)

        user_data = f"""#cloud-config
users:
  - name: agentsea
    ssh_authorized_keys:
      - {public_ssh_key}
    sudo: ['ALL=(ALL) NOPASSWD:ALL']
    groups: ['sudo']
    shell: /bin/bash
"""
        instance_type = self._choose_instance_type(cpu, memory)

        ssh_key_name = self._ensure_ssh_key(name, public_ssh_key)
        if not ssh_key_name:
            raise ValueError("SSH key name not provided or found")

        disk_size_gib = self._convert_disk_size_to_gib(disk)
        security_group_id = self._ensure_sg(
            "agentdesk-default", "agentdesk default vm sg"
        )

        tag_specifications = [
            {
                "ResourceType": "instance",
                "Tags": [
                    {"Key": "Name", "Value": name},
                    {"Key": "Owner", "Value": owner_id or "local"},
                    {"Key": "provisioner", "Value": "agentdesk"},
                ]
                + [{"Key": k, "Value": v} for k, v in (tags or {}).items()],
            }
        ]

        instances = self.ec2.create_instances(
            ImageId=image,  # type: ignore
            MinCount=1,
            MaxCount=1,
            InstanceType=instance_type,  # type: ignore
            KeyName=ssh_key_name,
            SecurityGroupIds=[security_group_id],
            BlockDeviceMappings=[
                {
                    "DeviceName": "/dev/sdh",
                    "Ebs": {"VolumeSize": disk_size_gib},
                }
            ],
            TagSpecifications=tag_specifications,  # type: ignore
            UserData=user_data,
        )
        instance_id = instances[0].id
        instances[0].wait_until_running()

        if reserve_ip:
            eip = boto3.client("ec2", region_name=self.region).allocate_address(
                Domain="vpc"
            )
            boto3.client("ec2", region_name=self.region).associate_address(
                InstanceId=instance_id, AllocationId=eip["AllocationId"]
            )

        instance = self.ec2.Instance(instance_id)
        public_ip = instance.public_ip_address

        # wait till agentd is ready
        self._wait_till_ready(public_ip, private_ssh_key=private_ssh_key)

        desktop = DesktopInstance(
            name=name,
            id=instance_id,
            addr=public_ip,
            cpu=cpu,
            memory=memory,  # type: ignore
            disk=disk,
            image=image,
            provider=self.to_data(),
            requires_proxy=True,
            owner_id=owner_id,
            metadata=metadata,
            key_pair_name=key_pair.name,
            ttl=ttl,
        )

        print(f"\nsuccessfully created desktop '{name}'")
        return desktop

    def _wait_till_ready(
        self,
        addr: str,
        local_agentd_port: Optional[int] = None,
        private_ssh_key: Optional[str] = None,
    ) -> None:
        if not local_agentd_port:
            local_agentd_port = find_open_port(8000, 9000)
            if not local_agentd_port:
                raise ValueError("could not find open port")
        print("waiting for desktop to be ready...")

        ready = False
        while not ready:
            print("waiting for desktop to be ready...")
            time.sleep(3)
            try:
                logger.debug("ensuring up ssh proxy...")
                pid = ensure_ssh_proxy(
                    local_port=local_agentd_port,
                    remote_port=8000,
                    ssh_host=addr,
                    ssh_key=private_ssh_key,
                    log_error=False,
                )
                atexit.register(cleanup_proxy, pid)

                logger.debug("calling agentd...")
                response = requests.get(f"http://localhost:{local_agentd_port}/health")
                logger.debug(f"agentd response: {response}")
                if response.status_code == 200:
                    ready = True
                cleanup_proxy(pid)
                atexit.unregister(cleanup_proxy)
            except Exception:
                try:
                    cleanup_proxy(pid, log_error=False)  # type: ignore
                except Exception:
                    pass

        logger.debug("cleaning up tunnel")
        try:
            cleanup_proxy(pid)  # type: ignore
            atexit.unregister(cleanup_proxy)
        except Exception:
            pass

    def _ensure_sg(self, name: str, description: str) -> str:
        # Attempt to find the default VPC
        vpcs = self.ec2_client.describe_vpcs(
            Filters=[{"Name": "isDefault", "Values": ["true"]}]
        )
        if not vpcs["Vpcs"]:
            raise Exception("No default VPC found in this region.")
        default_vpc_id = vpcs["Vpcs"][0]["VpcId"]  # type: ignore

        # Check if the security group already exists
        try:
            security_groups = self.ec2_client.describe_security_groups(
                Filters=[
                    {"Name": "group-name", "Values": [name]},
                    {"Name": "vpc-id", "Values": [default_vpc_id]},
                ]
            )
            if security_groups["SecurityGroups"]:
                # Security group already exists
                return security_groups["SecurityGroups"][0]["GroupId"]  # type: ignore
        except ClientError as e:
            print(f"Error checking for existing security group: {e}")

        # Security group does not exist, create it
        try:
            response = self.ec2_client.create_security_group(
                GroupName=name, Description=description, VpcId=default_vpc_id
            )
            security_group_id = response["GroupId"]

            # Add inbound rules
            self.ec2_client.authorize_security_group_ingress(
                GroupId=security_group_id,
                IpPermissions=[
                    {
                        "IpProtocol": "tcp",
                        "FromPort": 22,
                        "ToPort": 22,
                        "IpRanges": [{"CidrIp": "0.0.0.0/0"}],
                    },
                ],
            )
            return security_group_id
        except ClientError as e:
            raise Exception(f"Failed to create security group: {e}")

    def _convert_disk_size_to_gib(self, disk_size: str) -> int:
        """
        Converts a disk size specification string with units (e.g., "30gb", "1tb")
        to an integer representing the disk size in GiB.

        :param disk_size: Disk size string with units.
        :return: Disk size in GiB as an integer.
        """
        unit = disk_size[-2:].lower()
        size = int(disk_size[:-2])
        if unit == "gb":
            return size  # Assuming input in GiB, direct conversion for simplicity
        elif unit == "tb":
            return size * 1024  # Convert TB to GiB
        else:
            raise ValueError(f"Unsupported disk size unit: {unit}")

    def _choose_instance_type(self, cpu: int, memory: int) -> str:
        """
        Choose an EC2 instance type based on CPU and memory requirements.
        """
        # This is a simple mapping. Update it according to your needs.
        if cpu <= 2:
            if memory <= 4:
                return "t2.medium"
            elif memory <= 8:
                return "t2.large"
            else:
                return "t2.medium"
        elif cpu <= 4:
            if memory <= 16:
                return "t2.xlarge"
            else:
                return "t2.2xlarge"
        else:
            # Default to a larger instance for higher requirements
            return "t2.2xlarge"

    def _ensure_ssh_key(self, key_name: str, public_key_material: str) -> str:
        """
        Uploads an SSH public key to AWS EC2, if it does not already exist.
        """
        ec2_client = boto3.client("ec2", region_name=self.region)
        try:
            ec2_client.describe_key_pairs(KeyNames=[key_name])
            print(f"Key pair '{key_name}' already exists. Skipping import.")
            return key_name
        except ec2_client.exceptions.ClientError as e:
            if "InvalidKeyPair.NotFound" in str(e):
                ec2_client.import_key_pair(
                    KeyName=key_name, PublicKeyMaterial=public_key_material
                )
                print(f"Key pair '{key_name}' successfully imported.")
                return key_name
            raise

    def _get_ami_id_by_name(self, ami_name: str) -> str:
        """
        Find the latest custom AMI based on a specific naming pattern.

        Returns:
            The AMI ID of the latest custom AMI if found, otherwise None.
        """
        images = self.ec2_client.describe_images(
            Filters=[{"Name": "name", "Values": [ami_name]}]
        ).get("Images", [])
        if not images:
            raise ValueError(
                f"No images found with name: {ami_name} in region {self.region}"
            )
        return images[0]["ImageId"]  # type: ignore

    def _release_eip(self, instance: EC2Instance) -> None:
        # Assuming you have tagged your EIPs or have a way to associate them with instances
        filters = [{"Name": "instance-id", "Values": [instance.id]}]
        addresses = self.ec2_client.describe_addresses(Filters=filters)  # type: ignore
        for address in addresses.get("Addresses", []):
            self.ec2_client.release_address(AllocationId=address["AllocationId"])  # type: ignore
            print(f"Released EIP: {address['PublicIp']}")  # type: ignore

    def _delete_ssh_key(self, name: str) -> None:
        try:
            self.ec2_client.delete_key_pair(KeyName=name)
            print(f"Deleted SSH key: {name}")
        except self.ec2_client.exceptions.ClientError as e:
            print(f"Failed to delete SSH key {name}: {e}")

    def delete(self, name: str, owner_id: Optional[str] = None) -> None:
        instance = self._get_instance_by_name(name, owner_id=owner_id)
        if instance:
            # Release EIP if reserved for the instance
            # self._release_eip(instance) # TODO

            # Terminate the instance
            instance.terminate()
            instance.wait_until_terminated()
            print("Remote instance terminated")

            # TODO: for now we always create the key
            self._delete_ssh_key(name)

            # Remove the desktop VM from local state
            desk = DesktopInstance.get(name)
            if not desk:
                raise ValueError(
                    f"Desktop '{name}' not found in state, but deleted from provider"
                )
            desk.remove()

        keys = SSHKeyPair.find(owner_id=owner_id or "local")
        if keys:
            for key in keys:
                if (
                    "generated_for" in key.metadata
                    and key.metadata["generated_for"] == name
                ):
                    key.delete(key.name, key.owner_id)
                    print(f"Deleted SSH key {key.name}")

    def start(
        self,
        name: str,
        private_ssh_key: Optional[str] = None,
        owner_id: Optional[str] = None,
    ) -> None:
        desk = DesktopInstance.get(name, owner_id=owner_id)
        if not desk:
            raise ValueError(f"Desktop {name} not found")
        instance = self._get_instance_by_name(name)
        if instance:
            instance.start()
            instance.wait_until_running()

        if not instance:
            raise ValueError("Instance not found")

        public_ip = instance.public_ip_address
        self._wait_till_ready(public_ip, private_ssh_key=private_ssh_key)
        desk.addr = public_ip
        desk.status = "running"
        desk.save()

    def stop(self, name: str, owner_id: Optional[str] = None) -> None:
        desk = DesktopInstance.get(name, owner_id)
        if not desk:
            raise ValueError(f"Desktop {name} not found")
        instance = self._get_instance_by_name(name)
        if instance:
            instance.stop()
            instance.wait_until_stopped()
        desk.status = "stopped"
        desk.save()

    def list_remote(self) -> List[DesktopInstance]:
        instances = self.ec2.instances.filter(
            Filters=[{"Name": "instance-state-name", "Values": ["running", "stopped"]}]
        )
        desktops = []
        for instance in instances:
            desktops.append(DesktopInstance.load(instance.id))
        return desktops

    def list(self) -> List[DesktopInstance]:
        return DesktopInstance.find()

    def get_remote(self, name: str) -> Optional[DesktopInstance]:
        instance = self._get_instance_by_name(name)
        if not instance:
            return None
        return DesktopInstance.load(instance.id)

    def get(
        self, name: str, owner_id: Optional[str] = None
    ) -> Optional[DesktopInstance]:
        return DesktopInstance.get(name, owner_id=owner_id)

    def to_data(self) -> V1ProviderData:
        provider = V1ProviderData(type="ec2")
        if self.region:
            provider.args = {"region": self.region}
        return provider

    @classmethod
    def from_data(cls, data: V1ProviderData) -> EC2Provider:
        if data.args and "region" in data.args:
            region = data.args["region"]
            if region not in cls.AVAILABLE_REGIONS:
                raise ValueError(
                    f"Invalid region: {region}. Available regions are: "
                    f"{', '.join(sorted(cls.AVAILABLE_REGIONS))}"
                )
            return cls(region)
        return cls("us-east-1")

    def _get_instance_by_name(
        self, name: str, owner_id: Optional[str] = None
    ) -> Optional[EC2Instance]:
        filters = [{"Name": "tag:Name", "Values": [name]}]

        if owner_id is not None:
            filters.append({"Name": "Owner", "Values": [owner_id]})

        instances = self.ec2.instances.filter(Filters=filters)  # type: ignore

        return next((instance for instance in instances), None)

    def _get_root_device_size(self, instance: EC2Instance) -> str:
        for device in instance.block_device_mappings:
            if device.get("DeviceName") == instance.root_device_name:
                volume_id = device.get("Ebs", {}).get("VolumeId")
                if volume_id:
                    volume = boto3.resource("ec2").Volume(volume_id)
                    return f"{volume.size}gb"
        return "unknown"

    def refresh(self, log: bool = True) -> None:
        """Refresh state"""
        for vm in DesktopInstance.find():
            if not vm.provider:
                continue
            if vm.provider.type != "ec2":
                continue

            instance = self._get_instance_by_name(vm.name)
            if not instance:
                if log:
                    print(f"removing vm '{vm.name}' from state")
                vm.remove()
                return

            if not vm.reserved_ip:
                if vm.addr != instance.public_ip_address:
                    if log:
                        print(f"updating vm '{vm.name}' state")
                    vm.addr = instance.public_ip_address
                    vm.save()
                    return



import atexit
import base64
from datetime import datetime, timedelta
import json
import logging
import os
import random
import shutil
import signal
import string
import subprocess
import sys
import ssl
import httpx
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import shortuuid
from google.auth.transport.requests import Request
from google.cloud import container_v1
from google.oauth2 import service_account
from kubernetes import client, config
from kubernetes.client.rest import ApiException
from namesgenerator import get_random_name
from pydantic import BaseModel
from tenacity import retry, stop_after_attempt, wait_fixed

from agentdesk.util import find_open_port
from .base import DesktopInstance, DesktopProvider, V1ProviderData

logger = logging.getLogger(__name__)

ENABLE_NETWORK_POLICY = os.getenv("ENABLE_NETWORK_POLICY", "false").lower() == "true"


class GKEOpts(BaseModel):
    cluster_name: str
    region: str
    service_account_json: str

class VPNOpts(BaseModel):
    config_file: str
    username_password: str

class LocalOpts(BaseModel):
    path: Optional[str] = os.getenv("KUBECONFIG", os.path.expanduser("~/.kube/config"))


class KubeConnectConfig(BaseModel):
    provider: Literal["gke", "local"] = "local"
    namespace: str = "default"
    gke_opts: Optional[GKEOpts] = None
    local_opts: Optional[LocalOpts] = None
    branch: Optional[str] = None


def gke_opts_from_env(
    gke_sa_json=os.getenv("GKE_SA_JSON"),
    cluster_name=os.getenv("CLUSTER_NAME"),
    region=os.getenv("CLUSTER_REGION"),
) -> GKEOpts:
    if not gke_sa_json:
        raise ValueError("GKE_SA_JSON not set")
    if not cluster_name:
        raise ValueError("CLUSTER_NAME not set")
    if not region:
        raise ValueError("CLUSTER_REGION not set")
    return GKEOpts(
        service_account_json=gke_sa_json,
        cluster_name=cluster_name,
        region=region,
    )


DP = TypeVar("DP", bound="KubernetesProvider")


class KubernetesProvider(DesktopProvider):
    """A provider of desktop virtual machines"""

    def __init__(self, cfg: Optional[KubeConnectConfig] = None, vpn: Optional[VPNOpts] = None) -> None:
        self.cfg = cfg or KubeConnectConfig()
        self.vpn = vpn
        self.kubeconfig = None
        if cfg.provider == "gke":
            opts = cfg.gke_opts
            if not opts:
                raise ValueError("GKE opts missing")
            self.connect_to_gke(opts)
        elif cfg.provider == "local":
            opts = cfg.local_opts
            if not opts:
                opts = LocalOpts()
            if opts.path:
                config.load_kube_config(opts.path)
                self.kubeconfig = opts.path
        else:
            raise ValueError("Unsupported provider: " + cfg.provider)

        self.core_api = client.CoreV1Api()
        self.networking_api = client.NetworkingV1Api()

        self.namespace = cfg.namespace

        self.subprocesses = []
        self.setup_signal_handlers()

        self.branch = cfg.branch

    def create(
        self,
        name: Optional[str] = None,
        image: Optional[str] = None,
        memory: int = 2,
        cpu: int = 2,
        disk: str = "30gb",
        tags: Optional[Dict[str, str]] = None,
        reserve_ip: bool = False,
        ssh_key_pair: Optional[str] = None,
        owner_id: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        enable_basic_auth: bool = False,
        password: Optional[str] = None,
        sub_folder: Optional[str] = None,
        id: Optional[str] = None,
        ttl: Optional[int] = None,
        assigned: Optional[float] = None,
        node_selector: Optional[Dict[str, str]] = None,
        tolerations: Optional[List[client.V1Toleration]] = None,
    ) -> DesktopInstance:
        """Create a Desktop

        Args:
            name (str, optional): Name of the desktop. Defaults to random generation.
            image (str, optional): Image of the desktop. Defaults to Ubuntu Jammy.
            memory (int): Memory allotment. Defaults to 4gb.
            cpu (int): CPU allotment. Defaults to 2.
            disk (str): Disk allotment. Defaults to 30gb.
            tags (List[str], optional): Tags to apply to the desktop. Defaults to None.
            reserve_ip (bool, optional): Reserve an IP address. Defaults to False.
            ssh_key_pair (str, optional): SSH key pair name to use. Defaults to None.
            owner_id (str, optional): Owner of the desktop. Defaults to None.
            metadata (Dict[str, Any], optional): Metadata to apply to the instance. Defaults to None.
            enable_basic_auth (bool, optional): Enable basic auth. Defaults to False.
            password (str, optional): Password to use. Defaults to None.
            sub_folder (str, optional): Subfolder to use. Defaults to None.
            id (str, optional): ID of the desktop. Defaults to None.
            ttl (int, optional): Time to live seconds for the desktop. Defaults to None.

        Returns:
            DesktopInstance: An instance
        """
        if reserve_ip:
            raise NotImplementedError("Reserving IP addresses is not supported yet")
        if ssh_key_pair:
            raise NotImplementedError("SSH key pairs are not supported yet")

        if not name:
            name = get_random_name("-")
            if not name:
                raise ValueError("Could not generate a random name")

        env_vars = {}

        if not id:
            id = shortuuid.uuid()

        basic_auth_password = password
        if not password:
            basic_auth_password = "".join(
                random.choice(string.ascii_letters + string.digits) for _ in range(24)
            )
        basic_auth_user = None
        if enable_basic_auth:
            basic_auth_user = id
            env_vars["CUSTOM_USER"] = id
            env_vars["PASSWORD"] = basic_auth_password

        if sub_folder:
            env_vars["SUBFOLDER"] = sub_folder

        if not image:
            image = "us-docker.pkg.dev/agentsea-dev/agentd/desktop-webtop:latest"
        secrets: List[client.V1Secret] = []
        if env_vars:
            # Create a secret for the environment variables
            print("creating secret...")
            secret: Optional[client.V1Secret] = self.create_secret(name, env_vars)
            secrets.append(secret)
            env_from = [
                client.V1EnvFromSource(
                    secret_ref=client.V1SecretEnvSource(name=secret.metadata.name)  # type: ignore
                )
            ]
        else:
            env_from = []

        # -------- VPN Secret (only if self.vpn present) ---------
        vpn_secret_name = None
        if self.vpn:
            vpn_secret_name = f"{self._get_pod_name(name)}-vpn"
            print("creating vpn secret …")
            vpn_secret_body = client.V1Secret(
                api_version="v1",
                kind="Secret",
                metadata=client.V1ObjectMeta(
                    name=vpn_secret_name,
                    namespace=self.namespace,
                    labels={"provisioner": "agentdesk"},
                ),
                string_data={
                    "config.ovpn": self.vpn.config_file,
                    "auth.txt":    self.vpn.username_password,
                },
                type="Opaque",
            )
            self.core_api.create_namespaced_secret(
                namespace=self.namespace, body=vpn_secret_body
            )
            secrets.append(vpn_secret_body)

        # Resource configurations as before
        resources = client.V1ResourceRequirements(
            requests={"memory": f"{memory}Gi", "cpu": cpu},
            limits={"memory": "4Gi", "cpu": "4"},
        )

        logger.debug("using resources: ", resources.__dict__)

        pod_name = self._get_pod_name(name)

        # Desktop Container configuration
        desktop = client.V1Container(
            name=name,
            image=image,
            ports=[
                client.V1ContainerPort(container_port=8000),
                client.V1ContainerPort(container_port=3000),
                client.V1ContainerPort(container_port=3001),
            ],
            resources=resources,
            security_context=client.V1SecurityContext(
                capabilities=client.V1Capabilities(drop=["NET_ADMIN", "NET_RAW"]),
            ),
            env_from=env_from,  # Using envFrom to source env vars from the secret
            env=[client.V1EnvVar(name="CONTAINER_NAME", value=name)],
            image_pull_policy="Always",
        )
        if self.vpn:
            # 3. vpn sidecar
            vpn_container = client.V1Container(
                name="vpn-sidecar",
                image="us-docker.pkg.dev/agentsea-dev/agentd/vpn-sidecar:latest",
                security_context=client.V1SecurityContext(
                    privileged=True,
                    capabilities=client.V1Capabilities(add=["NET_ADMIN"]),
                ),
                volume_mounts=[
                    client.V1VolumeMount(name="dev-net-tun", mount_path="/dev/net/tun"),
                    client.V1VolumeMount(name="vpn-secret",  mount_path="/vpn-secret", read_only=True),
                ],
                resources=client.V1ResourceRequirements(
                    requests={"cpu": "100m", "memory": "128Mi"},
                    limits={"cpu": "500m", "memory": "256Mi"},
                ),
            )

            volumes = [
                client.V1Volume(
                    name="dev-net-tun",
                    host_path=client.V1HostPathVolumeSource(
                        path="/dev/net/tun",
                        type="CharDevice"
                    )
                ),
                client.V1Volume(
                    name="vpn-secret",
                    secret=client.V1SecretVolumeSource(
                        secret_name=vpn_secret_name,
                        default_mode=0o600,
                    )
                )
            ]
            # Pod specification
            pod_spec = client.V1PodSpec(
                containers=[desktop, vpn_container],
                restart_policy="Never",
                volumes=volumes,
                automount_service_account_token=False,
                enable_service_links=False,
                node_selector=node_selector,
                tolerations=tolerations,
            )
        else:
            # Pod specification
            pod_spec = client.V1PodSpec(
                containers=[desktop],
                restart_policy="Never",
                automount_service_account_token=False,
                enable_service_links=False,
                node_selector=node_selector,
                tolerations=tolerations,
            )

        # Pod creation
        pod = client.V1Pod(
            api_version="v1",
            kind="Pod",
            metadata=client.V1ObjectMeta(
                name=pod_name,
                labels={
                    "provisioner": "agentdesk",
                    "app": pod_name,
                    "workload": "desktop",
                    "branch": self.branch if self.branch else "undefined",
                },
                annotations={
                    "owner": owner_id,
                    "desktop_name": name,
                    "openmeter.io/subject": owner_id,
                    "data.openmeter.io/desktop_id": id,
                    "data.openmeter.io/workload": "desktop",
                    "data.openmeter.io/branch": self.branch if self.branch else "undefined",
                },
            ),
            spec=pod_spec,
        )

        try:
            created_pod: client.V1Pod = self.core_api.create_namespaced_pod(  # type: ignore
                namespace=self.namespace, body=pod
            )
            print(f"Pod created with name '{pod_name}'")
            # print("created pod: ", created_pod.__dict__)
            # Update secret's owner reference UID to newly created pod's UID
            if secrets:
                print("updating secret refs...")

                if not created_pod.metadata:
                    raise ValueError("expected pod metadata to be set")
                for secret in secrets:
                    if not secret or not secret.metadata:
                        raise ValueError("expected secret metadata to be set or None secret")
                    secret.metadata.owner_references = [
                        client.V1OwnerReference(
                            api_version="v1",
                            kind="Pod",
                            name=pod_name,
                            uid=created_pod.metadata.uid,  # This should be set dynamically after pod creation
                        )
                    ]
                    self.core_api.patch_namespaced_secret(
                        name=secret.metadata.name,
                        namespace=self.namespace,
                        body=secret,  # type: ignore
                    )
                print("secret refs updated")
        except ApiException as e:
            print(f"Exception when creating pod: {e}")
            raise

        if ENABLE_NETWORK_POLICY:
            self.create_network_policy(name)

        # Now, create the Service
        service_name = pod_name

        service = client.V1Service(
            api_version="v1",
            kind="Service",
            metadata=client.V1ObjectMeta(
                name=service_name,
                labels={"provisioner": "agentdesk"},
            ),
            spec=client.V1ServiceSpec(
                selector={"app": pod_name},
                ports=[
                    client.V1ServicePort(
                        name="agentd",
                        port=8000,
                        target_port=8000,
                    ),
                    client.V1ServicePort(
                        name="vnc",
                        port=3000,
                        target_port=3000,
                    ),
                    client.V1ServicePort(
                        name="vnc-https",
                        port=3001,
                        target_port=3001,
                    ),
                ],
                type="ClusterIP",
            ),
        )

        try:
            created_service = self.core_api.create_namespaced_service(
                namespace=self.namespace, body=service
            )
            print(f"Service created with name '{service_name}'")
        except ApiException as e:
            print(f"Exception when creating service: {e}")
            # Optionally, delete the Pod if Service creation fails
            self.core_api.delete_namespaced_pod(
                name=pod_name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(grace_period_seconds=5),
            )
            raise

        self.wait_pod_ready(name)
        self.wait_for_http_200(service_name)

        instance = DesktopInstance(
            id=id,
            name=name,
            cpu=cpu,
            memory=memory,
            disk=disk,
            metadata=metadata,
            owner_id=owner_id,
            provider=self.to_data(),
            vnc_port=3000,
            vnc_port_https=3001,
            agentd_port=8000,
            requires_proxy=True,
            image=image,
            resource_name=pod_name,
            namespace=self.namespace,
            basic_auth_user=basic_auth_user,
            basic_auth_password=basic_auth_password,
            ttl=ttl,
            assigned=assigned,
        )

        return instance

    def patch_meta_owner(self, owner_id, pod_name) -> client.V1Pod:
        """
        Patch the metadata of a Kubernetes pod to update the owner annotation.

        Args:
            owner_id (str): The new owner ID to set in the annotations.
            pod_name (str): The name of the pod to patch.

        Returns:
            V1Pod: The updated pod object.

        Raises:
            ApiException: If the patch request fails.
        """

        patch = {"metadata": {"annotations": {"owner": owner_id}}}

        try:
            updated_pod = self.core_api.patch_namespaced_pod(
                name=self._get_pod_name(pod_name),
                namespace=self.namespace,
                body=patch,
            )
            print(
                f"Pod '{pod_name}' updated successfully with new owner '{owner_id}'.",
                flush=True,
            )
        except ApiException as e:
            print(f"Failed to update pod '{pod_name}' with Error: {e}", flush=True)
            raise

        return updated_pod

    def create_network_policy(self, name: str) -> None:
        """
        Creates a NetworkPolicy that restricts the pod's network access.
        It allows egress to the internet while denying access to the cluster's internal network.
        """
        pod_name = self._get_pod_name(name)
        policy = client.V1NetworkPolicy(
            metadata=client.V1ObjectMeta(
                name=pod_name,
                namespace=self.namespace,
                labels={"provisioner": "agentdesk"},
            ),
            spec=client.V1NetworkPolicySpec(
                pod_selector=client.V1LabelSelector(match_labels={"app": pod_name}),
                policy_types=["Egress"],
                egress=[
                    client.V1NetworkPolicyEgressRule(
                        to=[
                            client.V1NetworkPolicyPeer(
                                ip_block=client.V1IPBlock(
                                    cidr="0.0.0.0/0",
                                    _except=[
                                        # Exclude common private IP ranges
                                        "10.0.0.0/8",
                                        "172.16.0.0/12",
                                        "192.168.0.0/16",
                                        "100.64.0.0/10",
                                    ],
                                )
                            )
                        ]
                    )
                ],
            ),
        )
        try:
            self.networking_api.create_namespaced_network_policy(
                namespace=self.namespace, body=policy
            )
            print(f"NetworkPolicy created for pod '{pod_name}'")
        except ApiException as e:
            print(f"Failed to create NetworkPolicy: {e}")
            raise

    def delete(self, name: str, owner_id: Optional[str] = None) -> None:
        """Delete a desktop

        Args:
            name (str): Name of the desktop
            owner_id (str, optional): Owner of the desktop. Defaults to None
        """
        errors = []
        try:
            pod_name = self._get_pod_name(name)
            # Delete the pod
            self.core_api.delete_namespaced_pod(
                name=pod_name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(grace_period_seconds=5),
            )
            print(f"Successfully deleted pod: {pod_name}")
        except ApiException as e:
            print(f"Failed to delete pod '{pod_name}': {e}")
            errors.append(e)

        # Attempt to delete the secret
        try:
            pod_name = self._get_pod_name(name)
            self.core_api.delete_namespaced_secret(
                name=pod_name, namespace=self.namespace
            )
            print(f"Successfully deleted secret: {pod_name}")
            self.core_api.delete_namespaced_secret(
                name=(pod_name + '-vpn'), namespace=self.namespace
            )
            print(f"Successfully deleted secret: {(pod_name + '-vpn')}")
        except ApiException as e:
            if e.status == 404:
                print(f"Secret '{pod_name}' not found, skipping deletion.")
            else:
                print(f"Failed to delete secret '{pod_name}': {e}")
                errors.append(e)

        try:
            self.core_api.delete_namespaced_service(
                name=pod_name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(),
            )
            print(f"Successfully deleted service: {self._get_pod_name(name)}")
        except ApiException as e:
            if e.status == 404:
                print(
                    f"Service '{self._get_pod_name(name)}' not found, skipping deletion."
                )
            else:
                print(f"Failed to delete service '{self._get_pod_name(name)}': {e}")
                errors.append(e)

        try:
            self.networking_api.delete_namespaced_network_policy(
                name=pod_name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(),
            )
            print(f"Successfully deleted NetworkPolicy: {pod_name}")
        except ApiException as e:
            if e.status == 404:
                print(f"NetworkPolicy '{pod_name}' not found, skipping deletion.")
            else:
                print(f"Failed to delete NetworkPolicy '{pod_name}': {e}")
                errors.append(e)

        if errors:
            raise Exception(errors)

    def start(
        self,
        name: str,
        private_ssh_key: Optional[str] = None,
        owner_id: Optional[str] = None,
    ) -> None:
        """Start a desktop

        Args:
            name (str): Name of the desktop
            private_ssh_key (str, optional): SSH key to use. Defaults to use ~/.ssh/id_rsa.
        """
        raise NotImplementedError("start not implemented")

    def stop(self, name: str, owner_id: Optional[str] = None) -> None:
        """Stop a desktop

        Args:
            name (str): Name of the desktop
            owner_id (str, optional): Owner of the desktop. Defaults to None
        """
        raise NotImplementedError("stop not implemented")

    def list(self) -> List[DesktopInstance]:
        """List desktops

        Returns:
            List[DesktopInstance]: A list of desktops
        """
        desktops = DesktopInstance.find()

        out = []
        for desktop in desktops:
            if not desktop.provider:
                continue
            if desktop.provider.type == "kube":
                out.append(desktop)

        return out

    def get(
        self, name: str, owner_id: Optional[str] = None
    ) -> Optional[DesktopInstance]:
        """Get a desktop

        Args:
            name (str): Name of the desktop
            owner_id (str, optional): Owner of the desktop. Defaults to None
        """
        desktops = DesktopInstance.find(name=name, owner_id=owner_id)

        for desktop in desktops:
            if not desktop.provider:
                continue
            if desktop.provider.type == "kube":
                return desktop

        return None

    def to_data(self) -> V1ProviderData:
        """Convert to a ProviderData object

        Returns:
            ProviderData: ProviderData object
        """
        cfg = self.cfg
        cfg.gke_opts = None

        return V1ProviderData(
            type="kube",
            args={"cfg": cfg.model_dump_json()},
        )

    @classmethod
    def from_data(cls, data: V1ProviderData) -> "KubernetesProvider":
        """From provider data

        Args:
            data (ProviderData): Provider data
        """
        config = None
        if data.args:
            config = KubeConnectConfig.model_validate_json(data.args["cfg"])

            if data.type == "gke":
                config.gke_opts = gke_opts_from_env()

        return cls(cfg=config)

    def refresh(self, log: bool = True, stale_pod_timeout: int = 300) -> None:
        """Refresh state"""

        label_selector = "provisioner=agentdesk"
        if self.branch:
            label_selector += f",branch={self.branch}"

        running_pods = self.core_api.list_namespaced_pod(
            namespace=self.namespace, label_selector=label_selector
        ).items

        # Fetch the agent instances from the database
        db_instances = self.list()

        # Create a mapping of pod names to pods
        running_pods_map = {pod.metadata.name: pod for pod in running_pods}  # type: ignore
        print("running_pods", list(running_pods_map), flush=True)
        # Create a mapping of instance names to instances
        db_instances_map = {instance.name: instance for instance in db_instances}

        # Check for instances in the database that are not running as pods
        for instance_name, instance in db_instances_map.items():
            if self._get_pod_name(instance_name) not in running_pods_map:
                print(
                    f"Instance '{instance_name}' is in the database but not running. Removing from database."
                )
                instance.delete(force=True)

        # Check for pods that are running but not present in the database and older than 5 minutes
        now = datetime.now(tz=datetime.now().astimezone().tzinfo)  # Ensure timezone awareness
        for pod_name, pod in running_pods_map.items():
            # Strip "desk-" prefix to match DB instance naming if applicable
            instance_name = pod_name.removeprefix("desk-")
            if instance_name not in db_instances_map:
                start_time = pod.status.start_time
                if start_time is None:
                    continue  # Skip pods without a known start time
                age = now - start_time
                if age > timedelta(seconds=stale_pod_timeout):
                    print(f"Pod '{pod_name}' not in DB and older than 5 minutes. Deleting.")
                    try:
                        self.delete(instance_name)
                    except Exception as e:
                        print(f"Error deleting pod '{pod_name}': {e}")

        logger.debug(
            "Refresh complete. State synchronized between Kubernetes and the database."
        )

    @classmethod
    def connect_config_type(cls) -> Type[KubeConnectConfig]:
        return KubeConnectConfig

    def connect_config(self) -> KubeConnectConfig:
        return self.cfg

    @classmethod
    def connect(cls, cfg: KubeConnectConfig) -> "KubernetesProvider":
        return cls(cfg)

    @retry(stop=stop_after_attempt(15))
    def connect_to_gke(self, opts: GKEOpts) -> Tuple[client.CoreV1Api, str, str, dict]:
        """
        Sets up and returns a configured Kubernetes client (CoreV1Api) and cluster details.

        Returns:
            Tuple containing the Kubernetes CoreV1Api client object, the project ID, and the cluster name.
        """
        service_account_info = json.loads(opts.service_account_json)
        credentials = service_account.Credentials.from_service_account_info(
            service_account_info,
            scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )

        # Setup GKE client to get cluster information
        gke_service = container_v1.ClusterManagerClient(credentials=credentials)
        project_id = service_account_info.get("project_id")
        if not project_id or not opts.cluster_name or not opts.region:
            raise ValueError(
                "Missing project_id, cluster_name, or region in credentials or metadata"
            )

        logger.debug("K8s getting cluster...")
        cluster_request = container_v1.GetClusterRequest(
            name=f"projects/{project_id}/locations/{opts.region}/clusters/{opts.cluster_name}"
        )
        cluster = gke_service.get_cluster(request=cluster_request)

        # Configure Kubernetes client
        logger.debug("K8s getting token...")
        ca_cert = base64.b64decode(cluster.master_auth.cluster_ca_certificate)
        try:
            logger.debug("K8s refreshing token...")
            credentials.refresh(Request())
        except Exception as e:
            logger.error("K8s token refresh failed: ", e)
            raise e
        access_token = credentials.token
        logger.debug("K8s got token: ", access_token)

        cluster_name = opts.cluster_name

        kubeconfig = {
            "apiVersion": "v1",
            "kind": "Config",
            "clusters": [
                {
                    "name": cluster_name,
                    "cluster": {
                        "server": f"https://{cluster.endpoint}",
                        "certificate-authority-data": base64.b64encode(
                            ca_cert
                        ).decode(),
                    },
                }
            ],
            "contexts": [
                {
                    "name": cluster_name,
                    "context": {
                        "cluster": cluster_name,
                        "user": cluster_name,
                    },
                }
            ],
            "current-context": cluster_name,
            "users": [
                {
                    "name": cluster_name,
                    "user": {
                        "token": access_token,
                    },
                }
            ],
        }
        self.kubeconfig = kubeconfig

        config.load_kube_config_from_dict(config_dict=kubeconfig)
        v1_client = client.CoreV1Api()
        logger.debug("K8s returning client...")

        return v1_client, project_id, cluster_name, kubeconfig

    @retry(stop=stop_after_attempt(200), wait=wait_fixed(2))
    def wait_for_http_200(self, name: str, path: str = "/", port: int = 8000):
        """
        Waits for an HTTP 200 response from the specified path on the given pod.

        Parameters:
            name (str): The name of the pod.
            path (str): The path to query. Defaults to root '/'.
            port (int): The port on which the pod service is exposed. Defaults to 8000.

        Raises:
            RuntimeError: If the response is not 200 after the specified retries.
        """
        pod_name = self._get_pod_name(name)
        logger.debug(
            f"Checking HTTP 200 readiness for pod {pod_name} on path {path} and port: {port}"
        )
        print(f"Waiting for desktop {name} to be ready...")
        status_code, response_text = self.call(
            name=name, path=path, method="GET", port=port
        )
        if status_code != 200:
            logger.info(f"Received status code {status_code}, with text: {response_text}, retrying...")
            raise Exception(
                f"Pod {pod_name} at path {path} is not ready. Status code: {status_code}"
            )
        logger.debug(f"Pod {pod_name} at path {path} responded with: {response_text}")
        logger.debug(f"Pod {pod_name} at path {path} is ready with status 200.")
        print(f"Health check passed for desktop '{name}'")

    @retry(stop=stop_after_attempt(200), wait=wait_fixed(2))
    def wait_pod_ready(self, name: str) -> bool:
        """
        Checks if the specified pod is ready to serve requests.

        Parameters:
            name (str): The name of the pod to check.

        Returns:
            bool: True if the pod is ready, False otherwise.
        """
        try:
            pod_name = self._get_pod_name(name)
            pod = self.core_api.read_namespaced_pod(
                name=pod_name, namespace=self.namespace
            )
            conditions = pod.status.conditions  # type: ignore
            if conditions:
                for condition in conditions:
                    if condition.type == "Ready" and condition.status == "True":
                        return True
            print("Waiting for pod to be ready...")
            raise Exception(f"Pod {pod_name} is not ready")
        except ApiException as e:
            print(f"Failed to read pod status for '{pod_name}': {e}")
            raise

    @retry(stop=stop_after_attempt(15))
    def call(
        self,
        name: str,
        path: str,
        method: str,
        port: int = 8000,
        data: Optional[dict] = None,
        headers: Optional[dict] = None,
    ) -> Tuple[int, str]:
        data = data or {}
        headers = headers or {}

        workload_proxy_url = os.getenv("WORKLOAD_PROXY_URL")
        if workload_proxy_url is not None:
            print("Using workload proxy at", workload_proxy_url)
            client_cert = os.getenv("WORKLOAD_PROXY_CLIENT_CERT")
            client_key = os.getenv("WORKLOAD_PROXY_CLIENT_KEY")
            ca_cert = os.getenv("WORKLOAD_PROXY_CA_CERT")

            context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=ca_cert)
            context.load_cert_chain(certfile=client_cert, keyfile=client_key)
            workload_proxy_client = httpx.Client(verify=context)

            merged_headers = {
                **headers,
                "X-Pod-Name": name,
                "X-Namespace": self.cfg.namespace,
                "X-Port": str(port),
            }
        else:
            print("Using direct connection to workload service")
            workload_proxy_client = httpx.Client()
            merged_headers = headers
            workload_proxy_url = (
                f"http://{name}.{self.cfg.namespace}.svc.cluster.local:{port}"
            )

        json_data = None if method == "GET" else data
        query_parameters = ""
        if method == "GET" and data:
            query_parameters = "?" + "&".join([f"{k}={v}" for k, v in data.items()])

        url = f"{workload_proxy_url.rstrip('/')}/{path.lstrip('/')}" + query_parameters

        print("Method: ", method)
        print("URL: ", url)
        print("Headers: ", merged_headers)
        print("JSON Data: ", json_data)

        r = workload_proxy_client.request(
            method=method, url=url, headers=merged_headers, json=json_data
        )

        return r.status_code, r.text

    def setup_signal_handlers(self):
        signal.signal(signal.SIGINT, self.graceful_exit)
        signal.signal(signal.SIGTERM, self.graceful_exit)
        atexit.register(self.cleanup_subprocesses)

    def _register_cleanup(self, proc: subprocess.Popen):
        self.subprocesses.append(proc)

    def cleanup_subprocesses(self):
        for proc in self.subprocesses:
            if proc.poll() is None:  # Process is still running
                proc.terminate()
                try:
                    proc.wait(timeout=10)
                except subprocess.TimeoutExpired:
                    proc.kill()
        self.subprocesses = []  # Clear the list after cleaning up

    def graceful_exit(self, signum, frame):
        self.cleanup_subprocesses()
        sys.exit(signum)  # Exit with the signal number

    def requires_proxy(self) -> bool:
        """Whether this runtime requires a proxy to be used"""
        return True

    def _get_pod_name(self, name: str) -> str:
        """Get the pod name for the given name"""
        return f"desk-{name}"

    def proxy(
        self,
        name: str,
        local_port: Optional[int] = None,
        container_port: int = 3000,
        background: bool = True,
        owner_id: Optional[str] = None,
    ) -> Tuple[int, Optional[int]]:
        """Proxy the desktop port to localhost.

        Args:
            name (str): Name of the agent
            local_port (Optional[int], optional): Local port to proxy to. Defaults to None.
            container_port (int, optional): Container port. Defaults to 3000.
            background (bool, optional): Whether to run in the background. Defaults to True.
            owner_id (Optional[str], optional): Owner ID. Defaults to None.

        Returns:
            Tuple[int, Optional[int]]: The local port being forwarded, and optionally the PID of the subprocess.
        """
        if local_port is None:
            local_port = find_open_port(container_port, container_port + 1000)
            if not local_port:
                raise RuntimeError("Failed to find an open port")

        # Prepare environment variables for the subprocess
        env = os.environ.copy()
        if self.kubeconfig:
            # Write kubeconfig to a temporary file
            import tempfile

            import yaml

            temp_kubeconfig = tempfile.NamedTemporaryFile(delete=False, mode="w")
            if isinstance(self.kubeconfig, str) and os.path.isfile(self.kubeconfig):
                temp_kubeconfig.close()
                shutil.copyfile(self.kubeconfig, temp_kubeconfig.name)
            else:
                yaml.safe_dump(self.kubeconfig, temp_kubeconfig)
                temp_kubeconfig.close()

            env["KUBECONFIG"] = temp_kubeconfig.name
        print(env["KUBECONFIG"])
        print(f"setting up port forward proxy with pod name: {self._get_pod_name(name)} local_port: {local_port} container_port: {container_port} and namespace {self.namespace}", flush=True)
        cmd = f"kubectl port-forward pod/{self._get_pod_name(name)} {local_port}:{container_port} -n {self.namespace}"

        print("Executing command:", cmd)
        if background:
            print("Running in background")
            proc = subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=env,  # Pass the environment variables to the subprocess
            )
            print("Process PID:", proc.pid)
            self._register_cleanup(proc)
            # Store the PID of the process in the class
            return (local_port, proc.pid)  # Return the PID of the subprocess
        else:
            try:
                subprocess.run(cmd, shell=True, check=True, env=env)
                return (
                    local_port,
                    None,
                )  # No PID to return when not in background mode
            except subprocess.CalledProcessError as e:
                raise RuntimeError(f"Port forwarding failed: {e}")

    def logs(
        self,
        name: str,
        follow: bool = False,
        owner_id: Optional[str] = None,
    ) -> Union[str, Iterator[str]]:
        """
        Fetches the logs from the specified pod. Can return all logs as a single string,
        or stream the logs as a generator of strings.

        Parameters:
            name (str): The name of the pod.
            follow (bool): Whether to continuously follow the logs.
            owner_id (Optional[str]): The owner ID of the pod. If provided, it will be included in the log lines.

        Returns:
            Union[str, Iterator[str]]: All logs as a single string, or a generator that yields log lines.
        """
        try:
            return self.core_api.read_namespaced_pod_log(
                name=self._get_pod_name(name),
                namespace=self.namespace,
                follow=follow,
                pretty="true",
                _preload_content=False,  # Important to return a generator when following
            )
        except ApiException as e:
            print(f"Failed to get logs for pod '{name}': {e}")
            raise

    def create_secret(self, name: str, env_vars: dict) -> client.V1Secret:
        """
        Creates a Kubernetes Secret object to store environment variables.

        Parameters:
            name (str): The base name of the secret, usually related to the pod name.
            env_vars (dict): A dictionary containing the environment variables as key-value pairs.

        Returns:
            client.V1Secret: The created Kubernetes Secret object.
        """
        logger.debug("creating secret with envs: ", env_vars)
        secret = client.V1Secret(
            api_version="v1",
            kind="Secret",
            metadata=client.V1ObjectMeta(
                name=self._get_pod_name(name),
                namespace=self.namespace,
                # This ensures that the secret is deleted when the pod is deleted.
                labels={"provisioner": "agentdesk"},
            ),
            string_data=env_vars,
            type="Opaque",
        )
        try:
            self.core_api.create_namespaced_secret(
                namespace=self.namespace, body=secret
            )
            print(f"Secret created '{self._get_pod_name(name)}'")
            return secret
        except ApiException as e:
            print(f"Failed to create secret: {e}")
            raise

    def runtime_local_addr(self, name: str, owner_id: Optional[str] = None) -> str:
        """
        Returns the local address of agentd with respect to the runtime
        """
        instances = DesktopInstance.find(name=name, owner_id=owner_id)
        if not instances:
            raise ValueError(f"No instances found for name '{name}'")
        instance = instances[0]

        return f"http://{instance.name}.{self.namespace}.svc.cluster.local:8000"

    def clean(
        self,
        owner_id: Optional[str] = None,
    ) -> None:
        pods = self.core_api.list_namespaced_pod(
            namespace="default", label_selector="provisioner=agentdesk"
        )
        for pod in pods.items:
            try:
                self.core_api.delete_namespaced_pod(
                    name=pod.metadata.name,
                    namespace="default",
                    body=client.V1DeleteOptions(grace_period_seconds=5),
                )
                print(f"Deleted pod: {pod.metadata.name}")
            except ApiException as e:
                print(f"Failed to delete pod '{pod.metadata.name}': {e}")
